/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992-1998, 1999 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   maus2rfc: Wandeln von MausTausch nach RFC.
*/

#include "config.h"

#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include <assert.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "maus.h"
#include "srfcmail.h"
#include "srfcnews.h"
#include "groups.h"
#include "domains.h"
#include "msg_io.h"
#include "uogetopt.h"
#include "setvars.h"

#ifdef WANT_NNTP
#	include "nntp.h"
#endif

#define HEADLEN	1024
enum messagetyp { NO_MSG, MAIL_MSG, NEWS_MSG, IGNORE_MSG };
enum specialtyp { NO_SPECIAL, INFO_SPECIAL, LOG_SPECIAL };
enum messagetyp this_message;
enum specialtyp this_special;

/* ------------------------------------------------------------------------- */
metamsg_t meta;

int work_only;
int dont_use_nntp;
int dont_use_nntp_inquire;
int dont_use_smtp;
#ifdef WANT_NNTP
static BOOLEAN use_inquire=FALSE;
#endif

static BOOLEAN maus2rfc (const char *fname);
static int count_refs(char *look);
static int prepend_refs(int ab, char *look, metamsg_t *msg);

static int opt_only_convert=0;
static int opt_quark_deliver=0;
static FILE *new_status_f=NULL;


/* ------------------------------------------------------------------------- */

static void parse_cmdlog(uostr_t *neu, char *start, char **next)
{
	const char *rueckmeldungen="?!";
	BOOLEAN err=FALSE;
	char *dies=start;

	/* wir parsen alle Zeilen nach dem Kommando und sehen nach,
	 * ob da eine Fehlermeldung kommt. Das fngt auch den
	 * eher unwahrscheinlichen Fall
	 * :"Kommando"
	 * :!bemerkung
	 * :?Fehler
	 * ab.
	 */
	do
	{
		dies=strpbrk(dies,"\r\n");
		if (!dies)
			break;
		dies+=2; /* \r\n ueberspringen */
	} while (strchr(rueckmeldungen,*dies));
	*next=dies;
	if (!err || !opts.smart_LOG || !neu)
		return;
	dies=start;
	do
	{
		char c;
		char *line=dies;
		dies=strpbrk(dies,"\r\n");
		if (!dies)
			break;
		c=*dies;
		*dies=0;
		uostr_xadd_cstr(neu,line);
		uostr_xadd_mem(neu,"\r\n",2);
		*dies=c;
		if (c=='\r')
			dies++;
		if (*dies=='\n')
			dies++;
	} while (strchr(rueckmeldungen,*dies));
}

static void parse_msglog(uostr_t *neu, char *start, char **next)
{
	/* Das Format des Logfiles:
	 *  :#12345
	 *  :!unwichtiges Gelaber
	 *  :?Haben sie eine Zeitmaschine (Datum/Uhrzeit in der Zukunft)
	 *  :=A123@...
	 * oder
	 *  :#12345
	 *  :?Das duerfen sie nicht.
	 * Ob eine Nachricht akzeptiert wurde, kann man also bei der ersten
	 * Gleichheitszeichenzeile sagen.
	 */
	char *dies;
	const char *rueckmeldungen="?!=";
	char *nullp;
	char nullc='\0'; /* keep gcc happy */
	char fname[PATH_MAX];
	char *mausid=NULL;
	char *transferid=NULL;
	char *errtext=NULL;
	char *spruchtext=NULL;
	int ok;
	BOOLEAN err=FALSE;

	dies=start;

	ok=0; /* noch keine "="-Zeile bekommen. */

	do
	{
		nullp=strpbrk(dies,"\r\n");
		if (!nullp)
		{
			*next=dies+strlen(dies);
		}
		else
		{
			char *p;
			p=nullp+1;
			while (*p=='\r' || *p=='\n')
				p++;
			*next=p;
		}

		/* start+1 zeigt jetzt auf die ID im Infile, 
		 * next auf die nchste Zeile, dies auf diese Zeile 
		 */
		if (nullp)
		{
			nullc=*nullp;
			*nullp='\0';
		}
		switch(*dies)
		{
		case '=':
			mausid=XSTRDUP(dies+1);
			ok=1;
			break;
		case '?': /* ein Fehler */
			err=TRUE;
		case '!': /* irgendein Spruch. */
			{
				char **indirect=(*dies=='!') ? &spruchtext : &errtext;
				if (!*indirect)
					*indirect=concat("\t",dies+1,"\r\n",NULL);
				else
				{
					*indirect=concat(*indirect,"\t",dies+1,"\r\n",NULL);
					XNULLCHECK(*indirect);
				}
			}
			break;
		case '#':
			transferid=XSTRDUP(dies+1);
			break;
		}
		if (nullp)
			*nullp=nullc;
		dies=*next;

	} while (*dies && strchr(rueckmeldungen,*dies));

	if (err && opts.smart_LOG && neu)
	{
		dies=start;
		do
		{
			char c;
			char *line=dies;
			dies=strpbrk(dies,"\r\n");
			if (!dies)
				break;
			c=*dies;
			*dies=0;
			uostr_xadd_cstr(neu,line);
			uostr_xadd_mem(neu,"\r\n",2);
			*dies=c;
			if (c=='\r')
				dies++;
			if (*dies=='\n')
				dies++;
		} while (strchr(rueckmeldungen,*dies));
	}

	assert(transferid);

	/* Statusmeldungen ... */
	if (strchr(transferid,'@') && !mausid)
	{
		if (errtext)
		{
			char *str;
			char *p,*q;
			/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
			str=XMALLOC(strlen(errtext));
			for (p=errtext,q=str;*p;p++)
			{
				if (*p=='\t')
					*q++=' '; /* space */
				else if (*p!='\r')
					*q++=*p;
			}
			*q=0;
			msglog ("! %s Error (statusmsg) : %s", transferid,str);
			free(str);
		}
		if (spruchtext)
		{
			char *str,*p,*q;
			/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
			str=XMALLOC(strlen(spruchtext));
			for (p=spruchtext,q=str;*p;p++)
			{
				if (*p=='\t')
					*q++=' '; /* space */
				else if (*p!='\r')
					*q++=*p;
			}
			*q=0;
			msglog ("! %s Notice (statusmsg) : %s", transferid,str);
			free(str);
		}
		if (!spruchtext && !errtext)
			msglog ("! %s (statusmsg) ok", transferid);
		free (spruchtext);
		free (errtext);
		free (transferid);
		return;
	}

	if (!mausid && !errtext)
	{
		free (spruchtext);
		free (transferid);
		return;
	}

	sprintf (fname, "%s/%s/outgoing/%s", 
		opts.spoolbase, action, transferid);

	if (opts.workmode==GATE)
	{
		/* dirty hack fr Gatemodus: Da die Transfer-ID des Batchfiles
		 * als Message-ID mitgeschickt wurde, bekommen wir nun eine
		 * Fehlermeldung fr das Batchfile, und damit die Mglichkeit,
		 * es zu lschen.
		 * Es sei denn natrlich wir haben eine lange MAUS-ID in 
		 * eine kurze verwandelt gehabt ... dann senden wir die
		 * als Transfer-ID und knnen den Batch nicht lschen. 
		 * Nicht unbedingt ideal.
		 */
		if (!strchr(transferid,'@'))
		{
			free (spruchtext);
			free (transferid);
			free (mausid);
			free (errtext);
			unlink(fname);
			return;
		}
	}

	if (errtext && !mausid)
	{
		uostr_t cmd=UOSTR_INIT;
		char *str;
		char *q;
		char *p;
		int was_mail=0;
		struct stat st;

		/* Struktur mit den Daten der gescheiterten Nachricht */
		metamsg_t done;
		/* Alles schoen als leer markieren */
		memset(&done,0,sizeof(metamsg_t));

		/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
		str=XMALLOC(strlen(errtext));
		for (p=errtext,q=str;*p;p++)
		{
			if (*p=='\t')
				*q++=' '; /* space */
			else if (*p!='\r')
				*q++=*p;
		}
		*q=0;

		/* Einmal per Syslog */
		if (opts.workmode==USER)
			log (LOG_WARNING, "Message %s failed: %s", fname, str);
		/* und ins Logfile */
		msglog ("! [%s]: FAILED: %s", transferid,str);
		free(str);

		/* So, jetzt die Message bouncen. Aber vorsichtig: Im Gateway-
		 * betrieb darf das nur fr Mail geschehen, im Userbetrieb
		 * sollte es fr News und Mail passieren.
		 */
		was_mail=!strncmp(transferid,"mail",4);
		if (was_mail || opts.workmode==USER)
		{
			/* Struktur mit der Bouncemeldung. */
			static long id_counter=0;
			metamsg_t msg;
			char zwi[1024];
			char idzwi[256];
			char ozwi[2048];
			char *absender=NULL;
			
			memset(&msg,0,sizeof(metamsg_t));
			msg.is_mail=was_mail;

			sprintf(zwi,"Mauscon@%s",opts.mysitedomain);
			msg.env_from=XSTRDUP(zwi);

			msg.head_from=XMALLOC(sizeof(metalist_t));
			msg.head_from->next=NULL;
			msg.head_from->ext=NULL;
			msg.head_from->text=XSTRDUP(zwi);

			/* this is ok */
			msg.subject=XSTRDUP("Message failed");

			msg.date=uo_now();

			sprintf(idzwi,"mauscon-bounce.%ld.%ld@%s",(long) getpid(),
				++id_counter,opts.mysitedomain);
			msg.message_id=XSTRDUP(idzwi);

			sprintf(ozwi,
				"Your message was not accepted by %s, because:\r\n",
					opts.pollsitedomain);
			meta_add_to_body(&msg,ozwi);
			meta_add_to_body(&msg,errtext);
			if (spruchtext)
			{
				sprintf(ozwi,
					"\r\n%s stated the following:\r\n",
						opts.pollsitedomain);
				meta_add_to_body(&msg,ozwi);
				meta_add_to_body(&msg,spruchtext);
			}

			/* Jetzt den Body der Message anhngenA
			 * im Hchstfall 16Kb, mehr lohnt irgendwie nicht.
			 */
			{
				uostr_t bouncebuf=UOSTR_INIT;
				if (!rfc_read_message(fname,&bouncebuf) && bouncebuf.data)
				{
					if (strlen(bouncebuf.data)<16*1024)
					{
						meta_add_to_body(&msg,"Failed message follows:\r\n\r\n");
						meta_add_to_body(&msg,bouncebuf.data);
					}
					else
					{
						meta_add_to_body(&msg,
							"Failed message excedes 16K - not copied!\n");
					}
					if (was_mail)
						absender=rfc_find_header (bouncebuf.data, 
							"x-mausconnect-env-from");
					if (!absender)
						absender=rfc_find_header (bouncebuf.data, 
							"reply-to");
					if (!absender)
						absender=rfc_find_header (bouncebuf.data, 
							"sender");
					if (!absender)
						absender=rfc_find_header (bouncebuf.data, 
							"from");
				}
				else
				{
					log(LOG_WARNING, _("cannot read %s (in bounce)"),fname);
					meta_add_to_body(&msg,"Sorry - cannot read message!");
				}
				uostr_freedata(&bouncebuf);
			}
			if (absender)
			{
				msg.env_to=XSTRDUP(absender);
				msg.head_to=XMALLOC(sizeof(metalist_t));
				msg.head_to->next=NULL;
				msg.head_to->ext=NULL;
				msg.head_to->text=XSTRDUP(absender);
				srfcmail(opt_only_convert,0,&msg);
				meta_free_list(&msg.head_to);

				free(msg.subject);
				free(msg.env_to);
				msg.subject=XSTRDUP("Message failed, gatemasters copy");
			}
			else
			{
				free(msg.subject);
				msg.subject=XSTRDUP("Message failed, sender unknown, waking gatemaster");
			}
			msg.head_to=XMALLOC(sizeof(metalist_t));
			msg.head_to->next=NULL;
			if (absender)
				msg.head_to->ext=NULL;
			else
				msg.head_to->ext=XSTRDUP("Mauscon needs your help");
			msg.head_to->text=XSTRDUP(opts.gatemaster);
			msg.env_to=XSTRDUP(opts.gatemaster);
			srfcmail(opt_only_convert,0,&msg);
			uostr_freedata(&msg.body);
			meta_clean_message(&msg);
			free(absender);
			/* Bounce abgeschlossen */
		} /* soll gebounced werden? */
		uostr_xdup_cstrmulti(&cmd,
			opts.spoolbase,"/",action,"/failed/",transferid,0);
		uostr_x0(&cmd);
		if ((0==stat(fname,&st) || errno!=ENOENT)
			&& rename(fname,cmd.data))
		{
			/* we call this only in case the failed/ directory is 
			 * on another file system. This should not happen, but
			 * we need to get the file out of the way _anyway_,
			 * regardless how stupid the admin is.
			 */
			uostr_xdup_cstrmulti(&cmd,
				"mv ",opts.spoolbase, " ",
				opts.spoolbase,"/",action,"/failed/",transferid,0);
			uostr_x0(&cmd);
			system(cmd.data);
		}
		uostr_freedata(&cmd);
		free (spruchtext);
		free (transferid);
		free (mausid);
		free (errtext);
		return;
	}
	/* Nachricht war OK.
	 * Im Usermodus nun die ursprngliche Nachricht canceln 
	 */
	msglog ("  [%s]: sent to %s", transferid,opts.pollsite);
	if (opts.workmode==USER 
		&& *transferid=='m' && transferid[1]=='t' && isdigit(transferid[2])
		&& (work_only==ONLY_ALL || work_only==ONLY_NEWS))
	{
		uostr_t buf=UOSTR_INIT;
		char fn2[PATH_MAX];
		sprintf (fn2, "%s/%s/outgoing/%s", 
			opts.spoolbase, action, transferid+2);
		if (0==rfc_read_message(fn2,&buf))
		{
			char *h;
			char *id;
			char *ng;
			id=rfc_find_header(buf.data,"Message-ID");
			ng=rfc_find_header(buf.data,"Newsgroups");
			h=rfc_find_header(buf.data,"Sender");
			if (!h)
				h=rfc_find_header(buf.data,"From");
			if (ng && id && h && 
				!rfc_find_header(buf.data,"X-MausConnect-Env-To"))
			{
				metamsg_t msg;
				
				memset(&msg,0,sizeof(metamsg_t));
				if (!rfc_to_meta(&buf,&msg,0)) /* 0: is_news */
				{
					char *zwi;

					/* Control-Header */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cancel <>")+1);
					sprintf(zwi,"cancel <%s>",msg.message_id);
					free(msg.control);
					msg.control=XSTRDUP(zwi);
					msg.transfer_id=XSTRDUP(transferid);

					/* Subject */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cmsg cancel <>")+1);
					sprintf(zwi,"cmsg cancel <%s>",msg.message_id);
					if (msg.subject)
						free(msg.subject);
					msg.subject=zwi;

					/* Neue Message-ID jetzt erst vergeben */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cancel-")+1);
					sprintf(zwi,"cancel-%s",msg.message_id);
					free(msg.message_id);
					msg.message_id=zwi;

#define CANCELTEXT "original canceled by Mauscon"
					uostr_dup_mem(&msg.body,CANCELTEXT,sizeof(CANCELTEXT)-1);
					uostr_dup_mem(&msg.body,"\r\n",2);
					msg.body.data[msg.body.len]=0;

					if (msg.distribution)
					{
						free(msg.distribution);
						msg.distribution=NULL;
					}
					if (opts.usenet_dist_lokal)
					{
						msg.distribution=XSTRDUP(opts.usenet_dist_lokal);
						srfcnews(opt_only_convert,&msg);
					}
					meta_clean_message(&msg);
				}
			}
		}
		uostr_freedata(&buf);
	}
	/* Originaldatei zu news jetzt in jedem Fall lschen, auch wenn 
	 * cancel_original nicht gewnscht ist. Grund: es knnte man
	 * gewnscht gewesen sein - dann liegt es jetzt hier herum ... 
	 */
	if (opts.workmode==USER 
		&& *transferid=='m' && transferid[1]=='t' && isdigit(transferid[2])
		&& (work_only==ONLY_ALL || work_only==ONLY_NEWS))
	{
		char fn2[PATH_MAX];
		sprintf (fn2, "%s/%s/outgoing/%s", 
			opts.spoolbase, action, transferid+2);
		unlink(fn2);
	}

	{
		char *id=NULL;
		char *to=NULL;
		char *subject=NULL;
		FILE *df;
		df=fopen(fname,"r");
		if (df)
		{
			char buf[1024];
			while (fgets(buf,sizeof(buf),df))
			{
				char *p=strpbrk(buf,"\r\n");
				if (p) *p=0;
				if (!*buf)
					break;
#define MAGIC "message-id:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						id=XSTRDUP(p);
				}
#undef MAGIC
#define MAGIC "subject:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						subject=XSTRDUP(p);
				}
#undef MAGIC
#define MAGIC "x-mausconnect-env-to:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						to=XSTRDUP(p);
				}
#undef MAGIC
			}
			fclose(df);
			if (id && to)
			{
				uostr_t ov_fname=UOSTR_INIT;
				uostr_xdup_cstrmulti(&ov_fname,
					opts.spoolbase,"/",action,"/status/overview",0);
				uostr_x0(&ov_fname);
				df=fopen(ov_fname.data,"a");
				if (df)
				{
					fprintf(df,"%s %s\n",mausid,transferid);
					fclose(df);
					uostr_xdup_cstrmulti(&ov_fname,
						opts.spoolbase,"/",action,"/status/",transferid,".status",0);
					uostr_x0(&ov_fname);
					df=fopen(ov_fname.data,"w");
					if (df)
					{	
						time_t t=time(NULL);
						fprintf(df,"Message-ID: %s\n",id);
						fprintf(df,"To: %s\n",to);
						if (subject)
							fprintf(df,_("Subject: %s\n"),subject);
						/* End of Header */
						fprintf(df,"\n");
						fprintf(df,_("Versendet: %s\n"),ctime(&t));
						fclose(df);
					}
				}
				uostr_freedata(&ov_fname);
			}
		}
		free(to);
		free(id);
		free(subject);
	}

	if (opts.keep_sent)
	{
		uostr_t keep_fname=UOSTR_INIT;
		uostr_xdup_cstrmulti(&keep_fname,
			opts.spoolbase,"/",action,"/sent2maus/",transferid,0);
		uostr_x0(&keep_fname);
		if (rename(fname,keep_fname.data))
		{
			uostr_t cmd=UOSTR_INIT;
			uostr_xdup_cstrmulti(&cmd, "mv -f ",fname," ",0);
			uostr_xadd_uostr(&cmd,&keep_fname); /* keep_fname already \0-terminated */
			if (system(cmd.data))
				unlink(fname); /* last possibility */
			uostr_freedata(&cmd);
		}
		uostr_freedata(&keep_fname);
	}
	else
		unlink(fname);
	free (spruchtext);
	free (transferid);
	free (mausid);
	free (errtext);
}

static void
work_ren (uostr_t *bodybuf)
{
	char *this;
	char *next;
	char *p;
	char *nullp;
	char *str_end;
	unsigned char *body;
	char *oldname=NULL;
	char *newname=NULL;

	body=bodybuf->data;

	this=next=body;

	while (this && *this)
	{
		nullp=strpbrk(this,"\r\n");
		if (nullp)
		{
			p=nullp+1;
			while (*p=='\r' || *p=='\n')
				p++;
			next=p;
			str_end=nullp;
		}
		else {
			next=strlen(this)+this;
			str_end=next;
		}
		switch(*this)
		{
		case 'O':
			/* alter Name */
			oldname=XMALLOC(str_end-this);
			memcpy(oldname,this+1,str_end-this-1);
			oldname[str_end-this-1]=0;
			break;
		case 'N':
			/* neuer Name */
			newname=XMALLOC(str_end-this);
			memcpy(newname,this+1,str_end-this-1);
			newname[str_end-this-1]=0;
			break;
		case 'D':
			/* datum der Umbenennung */
			/* Verarbeitung der Umbenennung */
			if (oldname && newname) {
				grp_data_t *g;
				g=grp_suche_mausname(oldname, opts.pollsite);
				if (g) {
					grp_rename_mausgroup(oldname,newname,opts.pollsite);
				}
			}
			break;
		default:
			break;
		} /* switch */
		this=next;
	}

}

static void
work_log (uostr_t *bodybuf)
{
	char *this;
	char *next;
	char *p;
	char *nullp;
	char nullc='\0'; /* keep gcc happy */
	unsigned char *body;
	uostr_t neu=UOSTR_INIT;
	char *last_marker=NULL;
	BOOLEAN in_meldung=FALSE;

	body=bodybuf->data;


	this=next=body;

	while (this && *this)
	{
		nullp=strpbrk(this,"\r\n");
		if (nullp)
		{
			p=nullp+1;
			while (*p=='\r' || *p=='\n')
				p++;
			next=p;
		}
		else
			next=strlen(this)+this;
		switch(*this)
		{
		case '\"':
			last_marker=this;
			in_meldung=FALSE;
			parse_cmdlog(&neu,this,&next);
			break;
		case '#':
			last_marker=this;
			in_meldung=FALSE;
			parse_msglog(&neu,this,&next);
			break;
		case '$': /* Infofilequittungen */
			in_meldung=FALSE;
			last_marker=this;
			/* p=strchr(this,'='); */
			p = memchr (this, '=', next-this);
			if (p)
			{
				*p='\0';
				if (nullp)
				{
					nullc=*nullp;
					*nullp=0;
				}
				info_new_crc(this+1,strtol(p+1,NULL,10));
				if (nullp)
				{
					*nullp=nullc;
				}
				*p='=';
			}
			break;
		default:
			break;
		} /* switch */
		this=next;
	}

	if (opts.smart_LOG)
		uostr_xdup_uostr(bodybuf,&neu);
	uostr_freedata(&neu);
}

static int 
count_refs(char *look)
{
	char *p;
	int  count=0;
	while (*look)
	{
		while (*look && *look!='<')
			look++;
		p=strchr(look,'>');
		if (!p)
			break;

		count++;
		look=p+1;
		while (isspace(*look) || *look==',')
			look++;
	}
	return count;
}

static int 
prepend_refs(int ab, char *look, metamsg_t *msg)
{
	metalist_t *last=NULL;
	char *p;
	int  count=0;
	while (*look)
	{
		while (*look && *look!='<')
			look++;
		p=strchr(look,'>');
		if (!p)
			break;

		if (count>=ab || count==0)
		{
			metalist_t *new=NULL;	
			*p='\0';
			meta_add_list(&new,look+1,NULL,1);
			if (new)
			{
				if (!last)
				{
					new->next=msg->references;
					msg->references=new;
				}
				else
				{
					new->next=last->next;
					last->next=new;
				}
				last=new;
			}
			*p='>';
		}
		count++;
		look=p+1;
		while (isspace(*look) || *look==',')
			look++;
	}
	return count;
}

static const char *
status_name(const char st)
{
	switch(st)
	{
	case 'N': return _("not read"); break;
	case 'Z': return _("deffered"); break;
	case 'B': return _("answered"); break;
	case 'G': return _("read"); break;
	case 'W': return _("forwarded"); break;
	case 'M': return _("in MausNet"); break;
	case 'A': return _("arrived at destination bbs"); break;
	case 'Y': return _("arrived at gateway"); break;
	case 'K': return _("copied"); break;
	case 'T': return _("exported to user"); break;
	default: return _("unknown status"); break;
	}
}

static void
send_new_status(void)
{
	char fname[PATH_MAX];
	metamsg_t msg;
	uostr_t buf=UOSTR_INIT;

	fclose(new_status_f);
	sprintf(fname,"%s/%s/status/status.new",
		opts.spoolbase,action);

	if (rfc_read_message(fname,&buf))
		return; /* ??? was ist das? */
	
	memset(&msg,0,sizeof(metamsg_t));
	if (rfc_to_meta(&buf,&msg,1)) /* 1: is_mail */
		return;
	msg.is_mail=1;
	if (srfcmail(FALSE,0,&msg))
		unlink(fname);
	else 
		log(LOG_NOTICE,_("cannot send mailstatuschange to gatemaster"));
	uostr_freedata(&buf);
}

static void
work_incoming_status(metamsg_t *msg)
{
	char fname[PATH_MAX];
	char buf[1024];
	char *t_id=NULL;
	FILE *of;
	char timebufnow[20];
	uo_sec70_t s70;
	uo_datetime_t dt;

	sprintf(fname,"%s/%s/status/overview",opts.spoolbase,action);
	of=fopen(fname,"r");
	if (!of)
		return;

	while (fgets(buf,sizeof(buf),of))
	{
		char *p=strpbrk(buf,"\r\n");
		if (p) *p=0;
		p=strpbrk(buf,"\t ");
		if (!p)
			continue;
		*p++=0;
		/* yessir, i'm paranoid */
		while (isspace(*p))
			p++;
		if (!*p)
			continue;
		if (!strcasecmp(buf,msg->message_id))
		{
			t_id=XSTRDUP(p);
			break;
		}
	}
	fclose(of);

	if (!t_id)
		return;

	sprintf(fname,"%s/%s/status/%s.status",
		opts.spoolbase,action,t_id);
	free(t_id);

	of=fopen(fname,"a+");
	if (!of)
	{
		free(t_id);
		return;
	}

	s70=uo_now();
	uo_sec702dt(&dt,&s70);
	fprintf(of,"status change received %s",
		yyyy_mm_dd_hh_mm_ss(timebufnow,&dt,'-'));

	if (msg->maus_status[1])
	{
		char timebuf2[20];
		maus_get_date(&s70,msg->maus_status+1);
		uo_sec702dt(&dt,&s70);
		fprintf(of,"status changed %s\n",yyyy_mm_dd_hh_mm_ss(timebuf2,&dt,'-'));
	}

	fprintf(of,_("new state; %s (%c)"),
		status_name(msg->maus_status[0]),msg->maus_status[0]);
	fputs("\n\n\n",of);

	if (opts.mail_status_change)
	{
		rewind(of); /* von nun an lesen wir dort */


		if (!new_status_f)
		{
			char *tmp;
			sprintf(fname,"%s/%s/status/status.new",
				opts.spoolbase,action);
			new_status_f=fopen(fname,"a");
			if (!new_status_f)
			{
				fclose(of);
				return;
			}

			tmp=concat("Mauscon","@",opts.mysitedomain,NULL);
			XNULLCHECK(tmp);
			fprintf(new_status_f,"X-Mausconnect-Env-To: <%s>\n",opts.gatemaster);
			fprintf(new_status_f,"X-Mausconnect-Env-From: <%s>\n",tmp);
			fprintf(new_status_f,"From: <%s>\n",tmp);
			fprintf(new_status_f,"To: <%s>\n",opts.gatemaster);
			fprintf(new_status_f,"Message-Id: <NeueStatusMsgs.%lu@%s>\n",
				(unsigned long) time(NULL), opts.mysitedomain);

			free(tmp);
			fprintf(new_status_f,
				"Subject: status changes received at %s\n\n",timebufnow);
		}

		while (fgets(buf,sizeof(buf),of))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p) *p=0;

			fprintf(new_status_f,"%s\n",buf);
			if (!*buf)
				break;
		}

		if (msg->maus_status[1])
		{
			char bf[20];
			maus_get_date(&s70,msg->maus_status+1);
			uo_sec702dt(&dt,&s70);
			fprintf(new_status_f,
				"status changed %s\n",yyyy_mm_dd_hh_mm_ss(bf,&dt,'-'));
		}

		fprintf(new_status_f,
				_("new state: %s (%c)%s"),
				status_name(msg->maus_status[0]),msg->maus_status[0],
				"\n\n\n");
	}
	fclose(of);

}

static BOOLEAN
maus2rfc (const char *inputfilename)
{
	mio_t *chan;
	uostr_t errbuf=UOSTR_INIT;
	mio_retcode_t mr;

	mr=mio_maustausch.open(inputfilename,"r",&chan,&errbuf);
	if (mr!=MIO_OK) {
		log(LOG_ERR,_("cannot open %s: %s"), inputfilename, errbuf.data);
		return FALSE;
	}

	do
	{
		mr=chan->fn->import(chan,&meta);
		if (mr==MIO_EOF)
			break;

		if (mr==MIO_RETRY)
			continue; /* Log done */

		if (mr==MIO_ERROR)
		{
			log(LOG_CRIT,UOSTR_EMPTY0(&chan->errtext)?
				_("unknown error"):chan->errtext.data);
			chan->fn->close(chan);
			return FALSE;
		}

		if (meta.subject && !strncmp(meta.subject,"cmsg ",5)
			&& !meta.also_control && !meta.control)
		{
			/* das passiert z.B., wenn eine Control-Message in eine
			 * Sysopgruppe weitergeleitet wird.
			 * Und das Newssystem darf das dann nicht als control-
			 * message verstehen.
			 * Yep, this is a hack ...
			 */
			meta.subject[4]='_';
		}
#ifdef WANT_NNTP
		if ((meta.env_to || meta.newsgroups) && !meta.maus_special_id
			&& meta.message_id && meta.references && use_inquire
			&& opts.max_references>1)
		{
			char *ref;
			/* 
			 * Versuchen, die References zu verlngern, wenn nur 
			 * eine vorliegt.
			 */
			if (!meta.references->next)
				ref=nntp_inquire_get_headerentry(meta.references->text,"references");
			else 
				ref=NULL;

			if (ref)
			{
				int count;

				count=count_refs(ref);
				if (count)
				{
					int ab;
					ab=count-opts.max_references+2;
					if (ab<0)
						ab=0;
					prepend_refs(ab, ref, &meta);
				}
			}
		}
#if 0
		/* Re: behandlung */
		if (strnhalfcasecmp(meta->subject,"re: ",4) == 0) {
		}
        if (!strnhalfcasecmp(zwi,"re: ",4))
            has_re=1;
        if (meta->references && !has_re)
            OUTPUT3("Subject: Re: ",zwi,crlf);
        else if (!meta->references && has_re)
            /* re: weg bei Nachrichten ohne References */
            OUTPUT3("Subject: ",zwi+4,crlf);
        else
            OUTPUT3("Subject: ",zwi,crlf);
#endif
#endif
		if (meta.maus_special_id)
		{
			minfo_t *found;
			if (work_only==ONLY_ALL || work_only==ONLY_SPECIAL)
			{

				found=info_find(meta.maus_special_id);

				if (found)
					info_work_new_infofile(&meta,found);
				else if (0==strcasecmp(meta.maus_special_id,"ITI")) {
					/* unsere erste ITI? */
					info_work_new_infofile(&meta,found);
				} else if (strcasecmp(meta.maus_special_id,"REN")==0
					&& meta.body.len) {
					work_ren(&meta.body);
				} else if (0!=strcasecmp(meta.maus_special_id,"HEAD")
					&& meta.body.len) {
					if (!strcasecmp(meta.maus_special_id,"LOG"))
						work_log(&meta.body);
					if (meta.body.len)
						meta.env_to=XSTRDUP(opts.mailuserdomain);
				}
			}
		}
		/* NICHT else if (meta.env_to), da das zwischenzeitlich
		 * gesetzt worden sein knnte.
		 */
		if (meta.env_to)
		{
			/* Hack: Bei bestimmten Absendern (localpart oder full qualified)
			 * leeren Envelope erzeugen.
			 */
			if (opts.empty_envelope_for && meta.is_mail) {

				mc_list_t *l;
				for (l=opts.empty_envelope_for;l;l=l->next) {
					char *p=strchr(l->data,'@');
					if (p) {
						if (0==strcasecmp(l->data,meta.env_from)) {
							msglog("%s: converted envelope sender %s to <>",
								meta.message_id,meta.env_from);
							free(meta.env_from);
							meta.env_from=XSTRDUP("");
							break;
						}
					} else {
						size_t len=strlen(l->data);
						if (0==strncasecmp(l->data,meta.env_from,len)
							&& meta.env_from[len]=='@') {
							msglog("%s: converted envelope sender %s to <>",
								meta.message_id,meta.env_from);
							free(meta.env_from);
							meta.env_from=XSTRDUP("");
							break;
						}
					}
				}
			}
			/* mail */
			if (work_only==ONLY_ALL || work_only==ONLY_MAIL
				|| (work_only==ONLY_SPECIAL && meta.maus_special_id))
				srfcmail(opt_only_convert,opt_quark_deliver,&meta);

			if (opts.autostatus && !meta.maus_special_id && !opt_quark_deliver)
			{
				uostr_t lockfname=UOSTR_INIT;
				int fatal_err=0;
				FILE *status_f;
				uostr_xadd_cstrmulti(&lockfname,opts.spoolbase,"/",action,
					"/etc/LOCK",NULL);
				uostr_x0(&lockfname);
				while (lock_create (lockfname.data, &fatal_err))
				{
					if (fatal_err)
					{
						log(LOG_ERR,_("cannot lock %s: %s"),lockfname.data,
							strerror(errno));
						break;
					}
					sleep(5);
				}
				if (!fatal_err)
				{
					uostr_t fname=UOSTR_INIT;
					uostr_xadd_cstrmulti(&fname,opts.spoolbase,"/",action,
						"/etc/status.msg",NULL);
					uostr_x0(&fname);
					status_f=fopen(fname.data,"a");
					if (status_f)
					{
						char z[30];
						BOOLEAN ok=TRUE;
						if (opts.send_short_ids) {
							char *m=maus_long_to_short_id(meta.message_id);
							if (m)
								fprintf(status_f,"#%s\n",m);
							else
								ok=FALSE;
						} else {
							fprintf(status_f,"#%s\n",meta.message_id);
						}
						if (ok) {
							uo_sec70_t now=uo_now();
							fprintf(status_f,"B%c\n",opts.autostatus);
							maus_make_date(z,now);
							fprintf(status_f,"E%s\n",z);
						}
						fclose(status_f);
					}
					lock_delete(lockfname.data);
					uostr_freedata(&fname);
				}
				uostr_freedata(&lockfname);
			}
		}
		else if (meta.maus_status)
		{
			work_incoming_status(&meta);
		}
		else if (meta.newsgroups)
		{
			if (work_only==ONLY_ALL || work_only==ONLY_NEWS)
				srfcnews(opt_only_convert,&meta);
		}
	} while (mr==MIO_RETRY || mr==MIO_OK);
	chan->fn->close(chan);
	return TRUE;
}

static const char *opt_inputfilename="-";
static const char *opt_cnffname=SYSCONFDIR "/" CNFNAME;

static uogetopt_t myopts[]={
    {'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
    "Channel to use (old name)","CHANNEL"},
    {'C',"channel",UOGO_STRING,&action,0,
    "Channel to use","CHANNEL"},
    {'c',"cnffile",UOGO_STRING,&opt_cnffname,0,
    "Configuration file to use","FILE"},
    {'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&opt_cnffname,0,
    "Configuration file to use","FILE"},
	{  0,"dont-use-nntp", UOGO_FLAG, &dont_use_nntp, 1,
	"do not send messages using NNTP", NULL},
	{  0,"dont-use-nntp-inquire", UOGO_FLAG, &dont_use_nntp_inquire, 1,
	"do not complete References\n"
	"Usually maus2rfc tries to complete the references\n"
	"header. Use this option to disable this (NNTP may\n"
	"still be used for delivery", NULL},
	{  0,"dont-use-smtp", UOGO_FLAG, &dont_use_smtp, 1,
	"do not send messages using SMTP", NULL},

    {'i',"input",UOGO_STRING,&opt_inputfilename,0,
    "File to read input from.\n"
	"Usually stdin is used","FILE"},
	{'O',"only-convert", UOGO_FLAG, &opt_only_convert, 1,
	"don't send anything, but send to stdout\n"
	"use for testing, not for real work!", NULL},
	{  0,"only-mail", UOGO_FLAG, &work_only, ONLY_MAIL,
	"convert only mail\nWill not touch news or special things",NULL},
	{  0,"only-news", UOGO_FLAG, &work_only, ONLY_NEWS,
	"convert only news\nWill not touch mail or special things",NULL},
	{  0,"only-special", UOGO_FLAG, &work_only, ONLY_SPECIAL,
	"convert only specials\nWill not touch mail or news",NULL},

    {'q',"quark-deliver",UOGO_FLAG,&opt_quark_deliver,1,
    "Special mode for DU3.\nDo not use.",NULL},
    {'s',"set",UOGO_CALLBACK,&setvarfn,0,
    "override variables in the configuration file","VAR=VALUE"},

    { 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
    "Copyright (C) 1999 Uwe Ohse\n"
    PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
    "to the extent permitted by law.\n"
    "You may redistribute copies of " PACKAGE "\n"
    "under the terms of the GNU General Public License.\n"
    "For more information about these matters, see the file named COPYING.\n"
    "You may redistribute copies of " PACKAGE "\n",0},

    {0,0,0,0,0,0,0} /* --help and --version */
};


int
main (int argc, char **argv)
{
	int exitcode=0;
	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

    uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out,
      "usage: rfc2maus [options] [channelname]\n"
      "convert RFC messages to MausTausch\n",
      myopts,
      "\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");

	if (!action) {
		action=argv[1];
		argv++;
	}
    if (!action) {
        log (LOG_ERR, _("channel not given"));
        exit(2);
    }
    if (argv[1])
    {
        log (LOG_ERR, _("too many arguments"));
        exit (2);
    }

    /* configfile lesen */
    if (do_conf (opt_cnffname, action,
		setvarscount ? setvars : NULL,setvarscount))
        exit (1);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (!dom_init())
		exit(1);
	if (info_read_list())
		exit(1);
	if (!grp_load_list(TRUE))
		exit (1);

#if defined(WANT_NNTP)
	if (opts.newshost && !dont_use_nntp_inquire)
		use_inquire=nntp_inquire_open(opts.newshost);
#endif

	exitcode=!maus2rfc (opt_inputfilename);

#if defined(WANT_NNTP)
	if (opts.newshost && use_inquire)
		nntp_inquire_close();
#endif

	if (!grp_save_list())
		log(LOG_ERR, _("cannot save grouplist"));

	if (new_status_f)
		send_new_status();

	srfcmail_end_all();
	srfcnews_end_all();

	info_write_list();

	closelog ();
	close_msglog();

	exit (exitcode);
}

